# django
from django.conf import settings as _settings

from . import default_settings


class Settings(object):
    def __getattr__(self, name):
        try:
            return getattr(_settings, name)
        except AttributeError:
            pass
        return getattr(default_settings, name)


settings = Settings()





class BaseChecker(object):
    """Base class for checkers
    In common case you want redefine in child only two attributes:
        * `name` attribute
        * `get_value` method
    """
    key = None  # key for Redis
    key_template = '{rule}:{checker}:{value}:int'  # template for key

    settings = NotImplemented   # Django settings
    name = NotImplemented       # checker name

    def __init__(self, connection, request, rule_type, **kwargs):
        self.connection = connection                    # for get_attempts and incr
        self.request = request                          # for logging
        self.limit = self.get_limit(rule_type)          # for check
        self.key = self.get_key(request, rule_type, **kwargs)   # for redis connection

    def get_key(self, request, rule_type, **kwargs):
        """Generate key for Redis
        """
        value = self.get_value(request, **kwargs)
        if value is None:
            return
        return self.key_template.format(
            rule=rule_type,
            checker=self.name,
            value=value
        )

    def get_value(self, request):
        """Get key parameter for Redis key.
        IP-address, client id etc.
        """
        raise NotImplementedError

    def get_limit(self, rule_type):
        """Git limit from rule for current check
        """
        if rule_type in self.settings.BRUTEFORCE_LIMITS:
            rule = self.settings.BRUTEFORCE_LIMITS[rule_type]
        else:
            rule = self.settings.BRUTEFORCE_LIMITS['default']
        return getattr(rule, self.name)

    def incr(self):
        """Increment attempts count
        """
        if not self.key:
            return
        if self.connection.exists(self.key):
            self.connection.incr(self.key, 1)
        else:
            self.connection.set(self.key, 1)
            self.connection.expire(self.key, self.settings.BRUTEFORCE_TIMELIMIT * 60)

    def get_attempts(self):
        """Get current attempts count
        """
        if not self.key:
            return 0
        data = self.connection.get(self.key)
        if not data:
            return 0
        return int(data)

    def log(self):
        """Log limit activation
        """
        # TODO: add logging
        pass

    def get_error(self):
        """Generate error message
        """
        key = 'BRUTEFORCE_ERROR_{}'.format(self.name)
        template = getattr(self.settings, key, self.settings.BRUTEFORCE_ERROR)
        return template.format(
            name=self.name,
            limit=self.limit,
            timelimit=self.settings.BRUTEFORCE_TIMELIMIT,
        )

    def check(self):
        """Check. Just check.
        Return True if all is fine, False otherwise.
        """
        # if disabled
        if not self.key:
            return True
        if not self.limit:
            return True
        # get attempts
        count = self.get_attempts()
        # if first activation
        if count == self.limit:
            self.log()
        # compare with limit
        return count < self.limit


class UserChecker(BaseChecker):
    name = 'user'

    def get_value(self, request, user=None, **kwargs):
        if not user:
            if not request:
                return
            if not getattr(request, 'user'):
                return
            if not getattr(request, 'user'):
                return
            user = request.user
        if not user.is_authenticated():
            return
        return user.pk


class IPChecker(BaseChecker):
    name = 'ip'

    def get_value(self, request, **kwargs):
        if not request:
            return
        return request.META['REMOTE_ADDR']


class CSRFChecker(BaseChecker):
    name = 'csrf'

    def get_value(self, request, **kwargs):
        if not request:
            return
        return getattr(request.POST, self.settings.CSRF_COOKIE_NAME, None)


class FrequencyChecker(UserChecker):
    name = 'freq'

    def log(self):
        pass

    def incr(self):
        if not self.key:
            return
        if not self.limit:
            return
        self.connection.set(self.key, 0)
        self.connection.expire(self.key, self.limit)

    def check(self):
        if not self.key:
            return True
        if not self.limit:
            return True
        if not self.connection.exists(self.key):
            return True
        ttl = self.connection.ttl(self.key_freq)
        if ttl >= 0:
            return True
        self.log()
        return False



# coding=utf-8
from .utils import Rule
from . import checkers


CSRF_COOKIE_NAME = 'csrfmiddlewaretoken'

BRUTEFORCE_ERROR = 'Too many requests ({name}). Maximum {limit} per {timelimit} minutes.'
"""
Error message.
Allowed format args:
    * {name} -- checker name.
    * {limit} -- maximum requests for this checker.
    * {timelimit} -- `settings.BRUTEFORCE_TIMELIMIT` value.
"""

BRUTEFORCE_PROTECTION_ENABLED = True
"""
Allow you quick disable all checks.
If BRUTEFORCE_PROTECTION_ENABLED as `False` then `Attempt.check` always return `True`.
"""

BRUTEFORCE_TIMELIMIT = 10  # minutes
"""
Time to live for all attempts counters.
"""

# Rules:
#   * Key: rule name (first arg for `Attempt`)
#   * Value: limits for checkers (namedtuple)
BRUTEFORCE_LIMITS = {
    'default': Rule(
        user=100,       # max requests for one user by BRUTEFORCE_TIMELIMIT
        ip=300,         # max requests for one IP by BRUTEFORCE_TIMELIMIT
        csrf=50,        # max requests with one CSRF token by BRUTEFORCE_TIMELIMIT
        freq=0,         # max request frequency for client [seconds]
    ),
}

# List of checkers for `Attempt`:
BRUTEFORCE_CHECKERS = (
    checkers.CSRFChecker,
    checkers.FrequencyChecker,
    checkers.IPChecker,
    checkers.UserChecker,
)

# Redis settings:
BRUTEFORCE_REDIS_HOST = 'localhost'
BRUTEFORCE_REDIS_PORT = 6379
BRUTEFORCE_REDIS_DB = 0



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from setuptools import setup


setup(
    name='djbrut',
    version='0.8.1',

    author='orsinium',
    author_email='master_fess@mail.ru',

    description='Framework for views in big projects on Django.',
    long_description=open('README.rst').read(),
    keywords='djbrut django bruteforce protection security redis',

    packages=['djbrut'],
    requires=['django', 'redis'],

    url='https://github.com/orsinium/django-bruteforce-protection',
    download_url='https://github.com/orsinium/django-bruteforce-protection/tarball/master',

    license='GNU Lesser General Public License v3.0',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Plugins',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU Lesser General Public License v3 or later (LGPLv3+)',
        'Programming Language :: Python',
    ],
)



# -*- coding: utf-8 -*-
import redis
from .settings import settings

for checker in settings.BRUTEFORCE_CHECKERS:
    checker.settings = settings


class Attempt(object):
    """Check request by all possible checkers for specified rule type
    """
    error = None
    connection = redis.StrictRedis(
        host=settings.BRUTEFORCE_REDIS_HOST,
        port=settings.BRUTEFORCE_REDIS_PORT,
        db=settings.BRUTEFORCE_REDIS_DB,
    )

    def __init__(self, rule_type, request, **kwargs):
        self.checkers = []
        for checker in settings.BRUTEFORCE_CHECKERS:
            checker = checker(self.connection, request, rule_type, **kwargs)
            self.checkers.append(checker)

    def incr(self):
        """Increment counters for all checkers
        """
        for checker in self.checkers:
            checker.incr()

    def check(self, incr=True):
        """Check request by all checkers
        Return True if all is fine, False otherwise.
        """
        if not settings.BRUTEFORCE_PROTECTION_ENABLED:
            return True
        for checker in self.checkers:
            if not checker.check():
                self.error = checker.get_error()
                return False
        if incr:
            self.incr()
        return True
